
class ScrobbledArtist:
    """Represents an artist that has been scrobbled (ie: exists in the user's Last.fm library)"""

    def __init__(self, artist_name, playcount):
        self.artist_name = artist_name
        self.playcount = playcount

    def __eq__(self, other):
        return isinstance(other, ScrobbledArtist) \
                and self.artist_name == other.artist_name \
                and self.playcount == other.playcount

    def __repr__(self):
        return str(self.__dict__)



import unittest
from lastipy.spotify.library import add_albums_to_library
from spotipy import Spotify
from lastipy.spotify.album import SpotifyAlbum

class AddAlbumsToLibraryTest(unittest.TestCase):

    def test_adding_less_than_max_request(self):
        spotify = Spotify()
        spotify.current_user = unittest.mock.MagicMock({'id': 'testUser'})
        spotify.current_user_saved_albums_add = unittest.mock.MagicMock()    
        dummy_albums = [
            self._build_dummy_album('123'),
            self._build_dummy_album('456'),
            self._build_dummy_album('789')
        ]
        add_albums_to_library(spotify, dummy_albums)
        spotify.current_user_saved_albums_add.assert_called_with(['123', '456', '789'])


    def test_adding_more_than_max_request(self):
        spotify = Spotify()
        spotify.current_user = unittest.mock.MagicMock({'id': 'testUser'})
        spotify.current_user_saved_albums_add = unittest.mock.MagicMock()    
       
        dummy_albums = []
        for _ in range(150):
            dummy_albums.append(self._build_dummy_album('123'))
        
        expected_chunks = []
        for _ in range(3):
            chunk = []
            for _ in range(50):
                chunk.append('123')
            expected_chunks.append(chunk)

        add_albums_to_library(spotify, dummy_albums)

        for i in range(3):
            spotify.current_user_saved_albums_add.assert_called_with(expected_chunks[i])


    def _build_dummy_album(self, spotify_id):
        return SpotifyAlbum(spotify_id=spotify_id, album_type='album', release_date_precision='day', release_date='2000-01-01')


import unittest
from unittest.mock import MagicMock
from spotipy import Spotify
from lastipy.spotify.library import get_saved_tracks
from lastipy.track import Track


class GetSavedTracksTest(unittest.TestCase):

    def setUp(self):
        self.mock_spotify = Spotify()
        self.mock_spotify.current_user = MagicMock({'id': 'dummyUser'})

    def test_fetch_single_page(self):
        expected_saved_tracks = [Track(
            track_name='Penny Lane', artist='The Beatles', spotify_id='123456789')
        ]

        self.mock_spotify.current_user_saved_tracks = MagicMock()
        mock_saved_tracks_response = {
            'items': [{
                'track': {
                    'id': '123456789',
                    'name': 'Penny Lane',
                    'artists': [
                        {
                            'name': 'The Beatles'
                        }
                    ]
                }
            }]
        }
        self.mock_spotify.current_user_saved_tracks.side_effect = [
            mock_saved_tracks_response, {'items': []}]

        fetched_saved_tracks = get_saved_tracks(self.mock_spotify)

        self.assertCountEqual(fetched_saved_tracks, expected_saved_tracks)

    def test_fetch_multiple_pages(self):
        expected_saved_tracks = [Track(
            track_name='Penny Lane', artist='The Beatles', spotify_id='123456789'),
            Track(track_name='A Day in the Life',
                         artist='The Beatles', spotify_id='987654321')
        ]

        self.mock_spotify.current_user_saved_tracks = MagicMock()
        mock_saved_tracks_response_page_1 = {
            'items': [{
                'track': {
                    'id': '123456789',
                    'name': 'Penny Lane',
                    'artists': [
                        {
                            'name': 'The Beatles'
                        }
                    ]
                }
            }]
        }
        mock_saved_tracks_response_page_2 = {
            'items': [{
                'track': {
                    'id': '987654321',
                    'name': 'A Day in the Life',
                    'artists': [
                        {
                            'name': 'The Beatles'
                        }
                    ]
                }
            }]
        }
        self.mock_spotify.current_user_saved_tracks.side_effect = [
            mock_saved_tracks_response_page_1, mock_saved_tracks_response_page_2, {'items': []}]

        fetched_saved_tracks = get_saved_tracks(self.mock_spotify)

        self.assertCountEqual(fetched_saved_tracks, expected_saved_tracks)



import logging
from lastipy.spotify.search import search_for_tracks
from lastipy.recommendations.recommended_track import RecommendedTrack

def fetch_recommendations(spotify, track, limit):
    logging.info("Fetching up to " + str(limit) + " recommendations based on " + str(track) + " in Spotify")
    track_in_spotify = search_for_tracks(spotify=spotify, query=track.artist + " " + track.track_name)[0]
    # TODO this accepts up to 5 tracks, so this could be more sophisticated by using multiple top-listened-to tracks in a single call
    recommendations = spotify.recommendations(seed_tracks=[track_in_spotify.spotify_id], limit=limit)

    parsed_recommendations = []
    for recommendation in recommendations['tracks']:
        parsed_recommendations.append(RecommendedTrack(track_name=recommendation['name'], artist=recommendation['artists'][0]['name'], spotify_id=recommendation['id'], recommendation_rating=1)) 
    
    logging.info("Fetched " + str(len(parsed_recommendations)) + " track recommendations")
    return parsed_recommendations



import unittest
from unittest.mock import patch
from lastipy.lastfm.library import recent_tracks
from unittest.mock import Mock
from requests import HTTPError
from lastipy.track import Track


class RecentArtistsTest(unittest.TestCase):

    dummy_user = 'dummyUser'
    dummy_api_key = '123456789'

    @patch('lastipy.lastfm.library.recent_tracks.fetch_paginated_response')
    def test_fetch(self, mock_paginated_endpoint_fetch):
        expected_tracks = [
            Track(track_name='Strawberry Fields Forever', artist='The Beatles'),
            Track(track_name='Badge', artist='Cream'),
            Track(track_name='Black Dog', artist='Led Zeppelin')
        ]

        mock_paginated_endpoint_fetch.return_value = [
            {
                'recenttracks': {
                    'track': [
                        {
                            'name': 'Strawberry Fields Forever',
                            'artist': {
                                'name': 'The Beatles'
                            }
                        },
                        {
                            'name': 'Badge',
                            'artist': {
                                'name': 'Cream'
                            }
                        }
                    ],
                    '@attr': {
                        'totalPages': '2'
                    }
                }
            },
            {
                'recenttracks': {
                    'track': [
                        {
                            'name': 'Black Dog',
                            'artist': {
                                'name': 'Led Zeppelin'
                            }
                        }
                    ],
                    '@attr': {
                        'totalPages': '2'
                    }
                }
            }
        ]

        fetched_tracks = recent_tracks.fetch_recent_tracks(
            user=self.dummy_user, api_key=self.dummy_api_key)

        self.assertCountEqual(fetched_tracks, expected_tracks)




import unittest
from lastipy.lastfm.library import period
from lastipy.lastfm.library.top_track import TopTrack
from unittest.mock import patch, Mock
from requests import HTTPError
from lastipy.lastfm.library.top_tracks import fetch_top_tracks


class TopTracksFetcherTest(unittest.TestCase):

    dummy_user = 'dummyUser'
    dummy_api_key = '123456789'

    @patch('lastipy.lastfm.library.top_tracks.fetch_paginated_response')
    def test_one_page_of_results(self, mock_paginated_endpoint_fetch):
        expected_track_1 = TopTrack(
            track_name="Stayin Alive", artist="Bee Gees", playcount=2)
        expected_track_2 = TopTrack(
            track_name="Badge", artist="Cream", playcount=7
        )

        json_response_page_1 = {
            'toptracks': {
                'track': [
                    {
                        'name': 'Stayin Alive',
                        'artist': {
                            'name': 'Bee Gees'
                        },
                        'playcount': '2'
                    }
                ]
            }
        }

        json_response_page_2 = {
            'toptracks': {
                'track': [
                    {
                        'name': 'Badge',
                        'artist': {
                            'name': 'Cream'
                        },
                        'playcount': '7'
                    }
                ]
            }
        }

        mock_paginated_endpoint_fetch.return_value = [
            json_response_page_1, json_response_page_2]

        fetched_tracks = fetch_top_tracks(user=self.dummy_user,
                                          api_key=self.dummy_api_key, a_period=period.ONE_MONTH)
        self.assertCountEqual(
            fetched_tracks, [expected_track_1, expected_track_2])

    @patch('lastipy.lastfm.library.top_tracks.fetch_paginated_response')
    def test_songs_with_one_playcount_ignored(self, mock_paginated_endpoint_fetch):
        json_response_page = {
            'toptracks': {
                'track': [
                    {
                        'name': 'Stayin Alive',
                        'artist': {
                            'name': 'Bee Gees'
                        },
                        'playcount': '1'
                    }
                ]
            }
        }

        mock_paginated_endpoint_fetch.return_value = [
            json_response_page]

        fetched_tracks = fetch_top_tracks(user=self.dummy_user,
                                          api_key=self.dummy_api_key, a_period=period.ONE_MONTH)
        self.assertCountEqual(fetched_tracks, [])



import logging
import requests
from requests import RequestException

MAX_RESULTS_PER_PAGE = 200
MAX_RETRIES = 10


def fetch_paginated_response(url, user, api_key, json_array_key, extra_request_params=None):
    """Fetches and returns an array of JSON responses from the given URL. Many of Last.fm's endpoints behave the same way by dividing
    results into "pages" of response objects. The only real difference between these endpoints is the type of response object returned,
    which can be specified here by the json_array_key parameter."""

    logging.debug("Fetching from " + url)
    paginated_json_responses = []
    page = 1
    total_pages = 1
    retries = 0
    while page <= total_pages:
        try:
            json_response = _send_request(
                url, _build_json_payload(user, api_key, page, extra_request_params))
            logging.debug("Response: " + str(json_response))
            paginated_json_responses.append(json_response)
            total_pages = int(
                json_response[json_array_key]['@attr']['totalPages'])
            page = page + 1
        except RequestException:
            if retries < MAX_RETRIES:
                logging.warning(
                    "Failed to fetch page " + str(page) + ". Retrying...")
                retries = retries + 1
            else:
                logging.warning("Failed to fetch page " + str(page) +
                                " after " + str(retries) + " retries. Giving up and moving on...")
                break

    return paginated_json_responses


def _send_request(url, json_payload):
    response = requests.get(url, params=json_payload)
    if response.ok:
        return response.json()
    else:
        response.raise_for_status()


def _build_json_payload(user, api_key, page, extra_request_params=None):
    payload = {
        'user': user,
        'format': 'json',
        'api_key': api_key,
        'limit': MAX_RESULTS_PER_PAGE,
        'page': page
    }

    if extra_request_params is not None:
        for param in extra_request_params:
            payload[param['key']] = param['value']

    return payload



import unittest
from unittest.mock import patch

from lastipy.recommendations.recommendations import generate_recommendations
from lastipy.lastfm.library.top_track import TopTrack
from lastipy.recommendations.recommended_track import RecommendedTrack


class RecommendationsGeneratorTest(unittest.TestCase):

    @patch('lastipy.recommendations.recommendations.fetch_recent_tracks')
    @patch('lastipy.recommendations.recommendations.fetch_top_tracks')
    @patch('lastipy.recommendations.recommendations.lastfm_recommendations')
    @patch('lastipy.recommendations.recommendations.calculate_ratings')
    def test_recent_tracks_are_filtered(self, mock_calculate_ratings, mock_fetch_recommendations, mock_top_tracks, mock_recent_tracks):
        recent_track_1 = TopTrack(
            track_name="SWALBR", artist="Cream", playcount=1)
        recent_track_2 = TopTrack(
            track_name="Badge", artist="Cream", playcount=1)
        recent_track_3 = TopTrack(
            track_name="Layla", artist="Derek & the Dominos", playcount=8)
        mock_recent_tracks.return_value = [
            recent_track_1, recent_track_2, recent_track_3]

        mock_top_tracks.return_value = [recent_track_3]

        new_recommendation = RecommendedTrack(
            track_name="Key to the Highway", artist="Derek & the Dominos", recommendation_rating=8)
        already_scrobbled_recommendation_1 = RecommendedTrack(
            track_name="Badge", artist="Cream", recommendation_rating=8)
        already_scrobbled_recommendation_2 = RecommendedTrack(
            track_name="SWALBR", artist="Cream", recommendation_rating=8)
        recommendations = [
            new_recommendation, already_scrobbled_recommendation_1, already_scrobbled_recommendation_2]
        mock_fetch_recommendations.fetch_recommendations.return_value = recommendations

        mock_calculate_ratings.return_value = recommendations

        self.assertCountEqual(generate_recommendations(spotify=None,
            user='test', lastfm_api_key=''), [new_recommendation])

    @patch('lastipy.recommendations.recommendations.fetch_recent_tracks')
    @patch('lastipy.recommendations.recommendations.fetch_top_tracks')
    @patch('lastipy.recommendations.recommendations.lastfm_recommendations')
    @patch('lastipy.recommendations.recommendations.calculate_ratings')
    def test_blacklisted_artists_are_filtered(self, mock_calculate_ratings, mock_fetch_recommendations, mock_top_tracks, mock_recent_tracks):
        mock_recent_tracks.return_value = []
        mock_top_tracks.return_value = [TopTrack(
            track_name='Here Comes the Sun', artist='The Beatles', playcount=5)]

        recommendation_1 = RecommendedTrack(
            track_name="Badge", artist="Cream", recommendation_rating=1)
        recommendation_2 = RecommendedTrack(
            track_name="Penny Lane", artist="The Beatles", recommendation_rating=1)
        recommendations = [recommendation_1, recommendation_2]
        mock_fetch_recommendations.fetch_recommendations.return_value = recommendations

        mock_calculate_ratings.return_value = recommendations

        recommendations = generate_recommendations(
            user="meeee", lastfm_api_key='', spotify=None, blacklisted_artists=['The Beatles'])

        self.assertCountEqual(recommendations, [recommendation_1])

    @patch('lastipy.recommendations.recommendations.fetch_recent_tracks')
    @patch('lastipy.recommendations.recommendations.fetch_top_tracks')
    @patch('lastipy.recommendations.recommendations.lastfm_recommendations')
    @patch('lastipy.recommendations.recommendations.calculate_ratings')
    def test_blacklisted_artists_filtering_should_ignore_case(self, mock_calculate_ratings, mock_fetch_recommendations, mock_top_tracks, mock_recent_tracks):
        mock_recent_tracks.return_value = []
        mock_top_tracks.return_value = [TopTrack(
            track_name='everything i wanted', artist='Billie Eilish', playcount=5)]

        recommendation = RecommendedTrack(
            track_name="Bad Music", artist="Zayn", recommendation_rating=1)
        recommendations = [recommendation]
        mock_fetch_recommendations.fetch_recommendations.return_value = recommendations

        mock_calculate_ratings.return_value = recommendations

        recommendations = generate_recommendations(
            user="", lastfm_api_key='', spotify=None, blacklisted_artists=['ZAYN'])

        self.assertEqual(0, len(recommendations))



import logging
import requests
from lastipy.lastfm.library.scrobbled_artist import ScrobbledArtist
from lastipy.lastfm.library.paginated_endpoint import fetch_paginated_response

URL = 'http://ws.audioscrobbler.com/2.0/?method=library.getartists'


def fetch_recent_artists(user, api_key):
    """Fetches recent artists for the given user"""

    logging.info("Fetching recent artists for user " + user)

    paginated_json_responses = fetch_paginated_response(
        URL, user, api_key, 'artists')

    artists = []
    for json_response in paginated_json_responses:
        for artist in json_response['artists']['artist']:
            artists.append(ScrobbledArtist(
                artist_name=artist['name'], playcount=int(artist['playcount'])))

    logging.info("Fetched " + str(len(artists)) + " artists")
    logging.debug("Fetched artists: " + str(artists))

    return artists



#!/usr/bin/env python3.7

from lastipy.spotify import library
from spotipy import Spotify
from lastipy.spotify import token
from configparser import ConfigParser
import argparse
import os
from lastipy import definitions
from lastipy.lastfm.library.top_tracks import fetch_top_tracks
from lastipy.lastfm.recommendations.similar_tracks import fetch_similar_tracks
from lastipy.lastfm.recommendations.recommendations import generate_recommendations
from lastipy.lastfm.library.recent_tracks import fetch_recent_tracks
from lastipy.lastfm.library.recent_artists import fetch_recent_artists
from lastipy.lastfm.library import period
from lastipy.track import Track
from numpy.random import choice
from spotipy import Spotify
from lastipy.spotify import token
from lastipy.util.setup_logging import setup_logging
import logging
from lastipy.util.parse_api_keys import ApiKeysParser
from lastipy.spotify import library, search, playlist
from lastipy.lastfm.library import track_info
from datetime import datetime, timezone
from dateutil.relativedelta import relativedelta
import iso8601

# TODO paramterize these?
NEW_FAVORITES_PLAYLIST = 'New Favorites'
OLD_FAVORITES_PLAYLIST = 'Old Favorites'
NEGLECTED_PLAYLIST = 'Neglected'


# TODO test
def organize_favorites():
    """Organizes a user's Spotify library. Saved tracks with more than args.saved_song_playount_limit plays are moved to a "New Favorites" playlist. 
       Tracks in the "New Favorites" playlist are moved to an "Old Favorites" playlist after they've been in "New Favorites" for 
       args.new_favorites_time_limit days, unless they have less than or equal to args.new_favorites_playcount_limit plays, in which case they're instead moved 
       to a "Neglected" playlist."""

    setup_logging('organize_favorites.log')
    args = _extract_args()
    spotify = Spotify(auth=token.get_token(
        args.spotify_user, args.spotify_client_id_key, args.spotify_client_secret_key))

    logging.info("Organizing " + args.spotify_user + "'s favorites")
    move_saved_tracks(spotify, args)
    move_new_favorites(spotify, args)
    logging.info("Done!")


def move_new_favorites(spotify, args):
    new_favorites_tracks = playlist.get_tracks_in_playlist(
        spotify, playlist_name=NEW_FAVORITES_PLAYLIST)
    tracks_to_move = []
    neglected_tracks = []
    for track in new_favorites_tracks:
        added_at = iso8601.parse_date(track.added_at)
        new_favorites_time_limit = added_at + \
            relativedelta(days=int(args.new_favorites_time_limit))
        if datetime.now(timezone.utc) >= new_favorites_time_limit:
            try:
                playcount = track_info.fetch_playcount(
                    track, args.lastfm_user, args.lastfm_api_key)
                if playcount <= int(args.new_favorites_playcount_limit):
                    neglected_tracks.append(track)
                else:
                    tracks_to_move.append(track)
            except:
                logging.warn("Couldn't get playcount for track " + str(track))

    logging.info("Moving " + str(len(tracks_to_move)) + " tracks from " +
                 NEW_FAVORITES_PLAYLIST + " to " + OLD_FAVORITES_PLAYLIST)
    playlist.remove_tracks_from_playlist(
        spotify, NEW_FAVORITES_PLAYLIST, tracks_to_move)
    playlist.add_tracks_to_playlist(
        spotify, OLD_FAVORITES_PLAYLIST, tracks_to_move)

    logging.info("Moving " + str(len(neglected_tracks)) +
                 " neglected tracks to " + NEGLECTED_PLAYLIST)
    playlist.remove_tracks_from_playlist(
        spotify, NEW_FAVORITES_PLAYLIST, neglected_tracks)
    playlist.add_tracks_to_playlist(
        spotify, NEGLECTED_PLAYLIST, neglected_tracks)


def move_saved_tracks(spotify, args):
    saved_tracks = library.get_saved_tracks(spotify)
    tracks_to_move = []
    for track in saved_tracks:
        try:
            playcount = track_info.fetch_playcount(
                track, args.lastfm_user, args.lastfm_api_key)
            if playcount >= int(args.saved_songs_playcount_limit):
                tracks_to_move.append(track)
        except:
            logging.warn("Couldn't get playcount for track " + str(track))

    logging.info("Moving " + str(len(tracks_to_move)) +
                 " tracks from library to " + NEW_FAVORITES_PLAYLIST)
    library.remove_tracks_from_library(spotify, tracks_to_move)
    playlist.add_tracks_to_playlist(
        spotify, NEW_FAVORITES_PLAYLIST, tracks_to_move)


def _extract_args():
    args_parser = _setup_args_parser()
    args = args_parser.parse_args()
    args = _extract_user_configs(args)
    args = _extract_api_keys(args)
    return args


def _setup_args_parser():
    parser = argparse.ArgumentParser(
        description="Organize Spotify saved tracks by removing tracks with a certain playcount from liked tracks and moving them to other playlists")
    parser.add_argument('user_configs_file',
                        type=argparse.FileType('r', encoding='UTF-8'))
    parser.add_argument(
        'api_keys_file', type=argparse.FileType('r', encoding='UTF-8'))
    return parser


def _extract_user_configs(args):
    config_parser = ConfigParser()
    config_parser.read(args.user_configs_file.name)
    section = 'Config'
    args.lastfm_user = config_parser[section]['LastFMUser']
    args.spotify_user = config_parser[section]['SpotifyUser']
    args.saved_songs_playcount_limit = config_parser[section]['SavedSongsPlaycountLimit']
    args.new_favorites_time_limit = config_parser[section]['NewFavoritesTimeLimit']
    args.new_favorites_playcount_limit = config_parser[section]['NewFavoritesPlaycountLimit']
    return args


def _str_to_bool(to_convert):
    return to_convert == "True"


def _extract_api_keys(args):
    keys_parser = ApiKeysParser(args.api_keys_file)
    args.lastfm_api_key = keys_parser.lastfm_api_key
    args.spotify_client_id_key = keys_parser.spotify_client_id_key
    args.spotify_client_secret_key = keys_parser.spotify_client_secret_key
    return args


if __name__ == "__main__":
    organize_favorites()
